S3_LANDING_BUCKET = "landing"
S3_ASSETS_BUCKET = "assets"
S3_FILE_STORE_BUCKET = "file-store"


S3_SOURCE_PATH_VAR = "S3_SOURCE_PATH"
S3_SOURCE_PATH_PREFIX_VAR = "S3_SOURCE_PATH_PREFIX"
S3_DESTINATION_PATH_VAR = "S3_DESTINATION_PATH"


# scratch file constants

S3_SCRATCH_KEY_PREFIX = "scratch/"
S3_SCRATCH_TAGGING_KEY = "time-to-live"
S3_SCRATCH_TAGGING_VALUE = "scratch"
